"use strict";
const tslib_1 = require("tslib");
const util_1 = tslib_1.__importDefault(require("util"));
const fs = tslib_1.__importStar(require("fs-extra"));
const path_1 = require("path");
const lisk_db_1 = require("@liskhq/lisk-db");
const semver = tslib_1.__importStar(require("semver"));
const command_1 = require("@oclif/command");
const cli_ux_1 = tslib_1.__importDefault(require("cli-ux"));
const constants_1 = require("./constants");
const client_1 = require("./client");
const config_1 = require("./utils/config");
const chain_1 = require("./utils/chain");
const events_1 = require("./events");
const genesis_block_1 = require("./utils/genesis_block");
const createAsset_1 = require("./createAsset");
const node_1 = require("./utils/node");
const path_2 = require("./utils/path");
const process_1 = require("./utils/process");
const block_1 = require("./utils/block");
const exception_1 = require("./utils/exception");
const commands_1 = require("./utils/commands");
const network_1 = require("./utils/network");
const fs_1 = require("./utils/fs");
const download_1 = require("./utils/download");
let configCoreV4;
class LiskMigrator extends command_1.Command {
    async run() {
        var _a, _b, _c, _d;
        const { flags } = this.parse(LiskMigrator);
        const liskCoreV3DataPath = (0, path_2.resolveAbsolutePath)((_a = flags['lisk-core-v3-data-path']) !== null && _a !== void 0 ? _a : constants_1.DEFAULT_LISK_CORE_PATH);
        const outputPath = (_b = flags.output) !== null && _b !== void 0 ? _b : (0, path_1.join)(__dirname, '..', 'output');
        const snapshotHeight = flags['snapshot-height'];
        const customConfigPath = flags.config;
        const autoMigrateUserConfig = (_c = flags['auto-migrate-config']) !== null && _c !== void 0 ? _c : false;
        const autoStartLiskCoreV4 = flags['auto-start-lisk-core-v4'];
        const pageSize = Number(flags['page-size']);
        const snapshotPath = flags['snapshot-path']
            ? (0, path_2.resolveAbsolutePath)(flags['snapshot-path'])
            : flags['snapshot-path'];
        const snapshotURL = flags['snapshot-url'];
        const network = flags.network;
        const useSnapshot = !!(snapshotPath || snapshotURL);
        if (network && !useSnapshot) {
            this.error('Either --snapshot-path= or --snapshot-url= must be provided when using --network=');
        }
        if (snapshotURL && (!snapshotURL.startsWith('http') || !snapshotURL.endsWith('tar.gz'))) {
            this.error(`Expected --snapshot-url to begin with http(s) and end with 'tar.gz' instead received ${snapshotURL}.`);
        }
        (0, path_2.verifyOutputPath)(outputPath);
        const networkIdentifier = await (0, network_1.getNetworkIdentifier)(network, liskCoreV3DataPath);
        const networkConstant = constants_1.NETWORK_CONSTANT[networkIdentifier];
        const outputDir = flags.output ? outputPath : `${outputPath}/${networkIdentifier}`;
        if (!fs.existsSync(outputDir))
            fs.mkdirSync(outputDir, { recursive: true });
        const filePathCommandsToExec = `${outputDir}/${constants_1.FILE_NAME.COMMANDS_TO_EXEC}`;
        const dataDir = (0, path_1.join)(__dirname, '..', constants_1.DEFAULT_DATA_DIR);
        try {
            if (useSnapshot) {
                if (snapshotURL === null || snapshotURL === void 0 ? void 0 : snapshotURL.startsWith('http')) {
                    cli_ux_1.default.action.start(`Downloading snapshot from ${snapshotURL} to ${outputDir}`);
                    await (0, download_1.downloadAndExtract)(snapshotURL, outputDir, dataDir);
                    cli_ux_1.default.action.stop(`Successfully downloaded snapshot from ${snapshotURL} to ${outputDir}`);
                }
                else if (snapshotPath === null || snapshotPath === void 0 ? void 0 : snapshotPath.endsWith('.tar.gz')) {
                    cli_ux_1.default.action.start(`Extracting snapshot to ${dataDir}`);
                    await (0, fs_1.extractTarBall)(snapshotPath, dataDir);
                    cli_ux_1.default.action.stop(`Successfully extracted snapshot to ${dataDir}`);
                }
            }
            else {
                const client = await (0, client_1.getAPIClient)(liskCoreV3DataPath);
                const nodeInfo = (await client.node.getNodeInfo());
                const { version: appVersion } = nodeInfo;
                cli_ux_1.default.action.start('Verifying if backup height from node config matches snapshot height');
                if (snapshotHeight !== nodeInfo.backup.height) {
                    this.error(`Lisk Core v3 backup height mismatch. Actual: ${nodeInfo.backup.height}, Expected: ${snapshotHeight}.`);
                }
                cli_ux_1.default.action.stop('Snapshot height matches backup height');
                cli_ux_1.default.action.start(`Verifying snapshot height to be multiples of round length i.e ${constants_1.ROUND_LENGTH}`);
                if (snapshotHeight % constants_1.ROUND_LENGTH !== 0) {
                    this.error(`Invalid snapshot height provided: ${snapshotHeight}. It must be an exact multiple of round length (${constants_1.ROUND_LENGTH}).`);
                }
                cli_ux_1.default.action.stop('Snapshot height is valid');
                (0, events_1.captureForgingStatusAtSnapshotHeight)(this, client, snapshotHeight, outputDir);
                if (autoStartLiskCoreV4) {
                    if (!networkConstant) {
                        this.error(`Unknown network detected. No NETWORK_CONSTANT defined for networkID: ${networkIdentifier}.`);
                    }
                }
                cli_ux_1.default.action.start('Verifying Lisk Core version');
                const isLiskCoreVersionValid = semver.valid(appVersion);
                if (isLiskCoreVersionValid === null) {
                    this.error(`Invalid Lisk Core version detected: ${appVersion}. Minimum supported version is ${constants_1.MIN_SUPPORTED_LISK_CORE_VERSION}.`);
                }
                if (semver.lt(appVersion, constants_1.MIN_SUPPORTED_LISK_CORE_VERSION)) {
                    this.error(`Lisk Migrator is not compatible with Lisk Core version ${appVersion}. Minimum supported version is ${constants_1.MIN_SUPPORTED_LISK_CORE_VERSION}.`);
                }
                cli_ux_1.default.action.stop(`${appVersion} detected`);
                await (0, chain_1.observeChainHeight)({
                    label: 'Waiting for snapshot height to be finalized',
                    liskCoreV3DataPath,
                    height: snapshotHeight,
                    delay: 500,
                    isFinal: true,
                });
            }
            await (0, chain_1.setTokenIDLskByNetID)(networkIdentifier);
            await (0, chain_1.setPrevSnapshotBlockHeightByNetID)(networkIdentifier);
            cli_ux_1.default.action.start('Creating database instance');
            const snapshotDirPath = await (0, path_2.resolveSnapshotPath)(useSnapshot, snapshotPath, dataDir, liskCoreV3DataPath);
            const db = new lisk_db_1.Database(snapshotDirPath);
            cli_ux_1.default.action.stop();
            cli_ux_1.default.action.start('Creating genesis assets');
            const createAsset = new createAsset_1.CreateAsset(db);
            const tokenID = (0, chain_1.getTokenIDLsk)();
            const genesisAssets = await createAsset.init(snapshotHeight, tokenID, pageSize);
            cli_ux_1.default.action.stop();
            const defaultConfigFilePath = await (0, config_1.resolveConfigPathByNetworkID)(networkIdentifier);
            const defaultConfigV4 = await fs.readJSON(defaultConfigFilePath);
            cli_ux_1.default.action.start(`Exporting genesis block to the path ${outputDir}`);
            await (0, genesis_block_1.writeGenesisAssets)(genesisAssets, outputDir);
            cli_ux_1.default.action.stop();
            if (autoMigrateUserConfig && !useSnapshot) {
                const configV3 = customConfigPath
                    ? await (0, config_1.getConfig)(this, liskCoreV3DataPath, networkIdentifier, customConfigPath)
                    : await (0, config_1.getConfig)(this, liskCoreV3DataPath, networkIdentifier);
                cli_ux_1.default.action.start('Creating backup for old config');
                await (0, config_1.createBackup)(configV3);
                cli_ux_1.default.action.stop();
                cli_ux_1.default.action.start('Migrating user configuration');
                const migratedConfigV4 = (await (0, config_1.migrateUserConfig)(configV3, defaultConfigV4, snapshotHeight));
                cli_ux_1.default.action.stop();
                cli_ux_1.default.action.start('Validating migrated user configuration');
                const isValidConfig = await (0, config_1.validateConfig)(migratedConfigV4);
                cli_ux_1.default.action.stop();
                if (!isValidConfig) {
                    throw new exception_1.MigratorException('Migrated user configuration is invalid.', 1);
                }
                cli_ux_1.default.action.start(`Exporting user configuration to the path: ${outputDir}`);
                await (0, config_1.writeConfig)(migratedConfigV4, outputDir);
                cli_ux_1.default.action.stop();
                configCoreV4 = migratedConfigV4;
            }
            cli_ux_1.default.action.start('Installing Lisk Core v4');
            await (0, node_1.installLiskCore)();
            cli_ux_1.default.action.stop();
            cli_ux_1.default.action.start('Creating genesis block');
            const blockHeaderAtSnapshotHeight = (await (0, block_1.getBlockHeaderByHeight)(db, snapshotHeight));
            await (0, genesis_block_1.createGenesisBlock)(this, networkConstant.name, defaultConfigFilePath, outputDir, blockHeaderAtSnapshotHeight);
            cli_ux_1.default.action.stop();
            cli_ux_1.default.action.start('Creating genesis block tar and SHA256 files');
            await (0, genesis_block_1.writeGenesisBlock)(outputDir);
            this.log(`Genesis block tar and SHA256 files have been created at: ${outputDir}.`);
            cli_ux_1.default.action.stop();
            if (!useSnapshot) {
                if (autoStartLiskCoreV4) {
                    try {
                        if (!autoMigrateUserConfig) {
                            configCoreV4 = defaultConfigV4;
                        }
                        cli_ux_1.default.action.start('Copying genesis block to the Lisk Core executable directory');
                        const liskCoreExecPath = await (0, process_1.execAsync)('which lisk-core');
                        const liskCoreV4ConfigPath = (0, path_1.resolve)(liskCoreExecPath, '../..', `lib/node_modules/lisk-core/config/${networkConstant.name}`);
                        await (0, genesis_block_1.copyGenesisBlock)(`${outputDir}/genesis_block.blob`, `${liskCoreV4ConfigPath}/genesis_block.blob`);
                        this.log(`Genesis block has been copied to: ${liskCoreV4ConfigPath}.`);
                        cli_ux_1.default.action.stop();
                        const isLiskCoreV3Stopped = await cli_ux_1.default.confirm("Please stop Lisk Core v3 to continue. Type 'yes' and press Enter when ready. [yes/no]");
                        if (isLiskCoreV3Stopped) {
                            let numTriesLeft = 3;
                            while (numTriesLeft) {
                                numTriesLeft -= 1;
                                const isCoreV3Running = await (0, node_1.isLiskCoreV3Running)(liskCoreV3DataPath);
                                if (!isCoreV3Running)
                                    break;
                                if (numTriesLeft >= 0) {
                                    const isStopReconfirmed = await cli_ux_1.default.confirm("Lisk Core v3 still running. Please stop the node, type 'yes' to proceed and 'no' to exit. [yes/no]");
                                    if (!isStopReconfirmed) {
                                        throw new Error(`Cannot proceed with Lisk Core v4 auto-start. Please continue manually. In order to access legacy blockchain information posts-migration, please copy the contents of the ${snapshotDirPath} directory to 'data/legacy.db' under the Lisk Core v4 data directory (e.g: ${constants_1.DEFAULT_LISK_CORE_PATH}/data/legacy.db/). Exiting!!!`);
                                    }
                                    else if (numTriesLeft === 0 && isStopReconfirmed) {
                                        const isCoreV3StillRunning = await (0, node_1.isLiskCoreV3Running)(liskCoreV3DataPath);
                                        if (isCoreV3StillRunning) {
                                            throw new Error(`Cannot auto-start Lisk Core v4 as Lisk Core v3 is still running. Please continue manually. In order to access legacy blockchain information posts-migration, please copy the contents of the ${snapshotDirPath} directory to 'data/legacy.db' under the Lisk Core v4 data directory (e.g: ${constants_1.DEFAULT_LISK_CORE_PATH}/data/legacy.db/). Exiting!!!`);
                                        }
                                    }
                                }
                            }
                            const isUserConfirmed = await cli_ux_1.default.confirm(`Start Lisk Core with the following configuration? [yes/no]
							${util_1.default.inspect(configCoreV4, false, 3)} `);
                            if (isUserConfirmed) {
                                cli_ux_1.default.action.start('Starting Lisk Core v4');
                                const networkName = networkConstant.name;
                                await (0, node_1.startLiskCore)(this, liskCoreV3DataPath, configCoreV4, networkName, outputDir);
                                this.log(`Started Lisk Core v4 at default data directory ('${constants_1.DEFAULT_LISK_CORE_PATH}').`);
                                cli_ux_1.default.action.stop();
                            }
                            else {
                                this.log('User did not accept the migrated config. Skipping the Lisk Core v4 auto-start process.');
                            }
                        }
                        else {
                            throw new Error(`User did not confirm Lisk Core v3 node shutdown. Skipping the Lisk Core v4 auto-start process. Please continue manually. In order to access legacy blockchain information posts-migration, please copy the contents of the ${snapshotDirPath} directory to 'data/legacy.db' under the Lisk Core v4 data directory (e.g: ${constants_1.DEFAULT_LISK_CORE_PATH}/data/legacy.db/). Exiting!!!`);
                        }
                    }
                    catch (err) {
                        const errorMsg = `Failed to auto-start Lisk Core v4.\nError: ${err.message}`;
                        throw new exception_1.MigratorException(errorMsg, err instanceof exception_1.MigratorException ? err.code : 3);
                    }
                }
                else {
                    this.log(`Please copy the contents of ${snapshotDirPath} directory to 'data/legacy.db' under the Lisk Core v4 data directory (e.g: ${constants_1.DEFAULT_LISK_CORE_PATH}/data/legacy.db/) in order to access legacy blockchain information.`);
                    this.log('Please copy genesis block to the Lisk Core V4 network directory.');
                }
            }
        }
        catch (error) {
            const commandsToExecute = [];
            const code = Number(`${error.code}`);
            const basicStartCommand = `lisk-core start --network ${networkConstant.name}`;
            const liskCoreStartCommand = (_d = (0, node_1.getLiskCoreStartCommand)()) !== null && _d !== void 0 ? _d : basicStartCommand;
            const backupLegacyDataDirCommand = `mv ${liskCoreV3DataPath} ${constants_1.LISK_V3_BACKUP_DATA_DIR}`;
            const copyLegacyDBCommand = `cp -r ${((0, path_1.resolve)(constants_1.LISK_V3_BACKUP_DATA_DIR, constants_1.SNAPSHOT_DIR), constants_1.LEGACY_DB_PATH)}`;
            if ([0, 1, 2].includes(code)) {
                const genesisBlockCreateCommand = (0, genesis_block_1.getGenesisBlockCreateCommand)();
                commandsToExecute.push('\n', '## Create the genesis block', '## NOTE: This requires installing Lisk Core v4 locally. Please visit https://lisk.com/documentation/lisk-core/v4/setup/npm.html for further instructions', '\n');
                commandsToExecute.push(genesisBlockCreateCommand);
                commandsToExecute.push('\n', '-----------------------------------------------------', '\n');
            }
            if ([
                0,
                1,
                2,
                4,
            ].includes(code)) {
                commandsToExecute.push('\n', '## Backup Lisk Core v3 data directory', '\n');
                commandsToExecute.push(backupLegacyDataDirCommand);
                commandsToExecute.push('\n', '-----------------------------------------------------', '\n');
            }
            if ([
                0,
                1,
                2,
                4,
                5,
            ].includes(code)) {
                commandsToExecute.push('\n', '## Copy legacy (v3) blockchain information to Lisk Core v4 legacy.db', '\n');
                commandsToExecute.push(copyLegacyDBCommand);
                commandsToExecute.push('\n', '-----------------------------------------------------', '\n');
            }
            if ([
                0,
                1,
                2,
                4,
                5,
                3,
            ].includes(code)) {
                commandsToExecute.push('\n', '## Lisk Core v4 start command - Please modify if necessary', '\n');
                commandsToExecute.push(liskCoreStartCommand);
                commandsToExecute.push('\n', '-----------------------------------------------------', '\n');
            }
            await (0, commands_1.writeCommandsToExec)(this, networkConstant, outputDir, commandsToExecute);
            this.error(`Migrator could not finish execution successfully due to: ${error.message}\nPlease check the commands to be executed in the file: ${filePathCommandsToExec}`);
        }
        await (0, commands_1.writeCommandsToExec)(this, networkConstant, outputDir);
        this.log('Successfully finished migration. Exiting!!!');
        process.exit(0);
    }
}
LiskMigrator.description = 'Migrate Lisk Core to latest version';
LiskMigrator.flags = {
    version: command_1.flags.version({ char: 'v' }),
    help: command_1.flags.help({ char: 'h' }),
    output: command_1.flags.string({
        char: 'o',
        required: false,
        description: `File path to write the genesis block. If not provided, it will default to cwd/output/{v3_networkIdentifier}/genesis_block.blob. Do not use any value starting with the default data path reserved for Lisk Core: '${constants_1.DEFAULT_LISK_CORE_PATH}'.`,
    }),
    'lisk-core-v3-data-path': command_1.flags.string({
        char: 'd',
        required: false,
        description: 'Path where the Lisk Core v3.x instance is running. When not supplied, defaults to the default data directory for Lisk Core.',
    }),
    config: command_1.flags.string({
        char: 'c',
        required: false,
        description: 'Custom configuration file path for Lisk Core v3.1.x.',
    }),
    'snapshot-height': command_1.flags.integer({
        char: 's',
        required: true,
        env: 'SNAPSHOT_HEIGHT',
        description: 'The height at which re-genesis block will be generated. Can be specified with SNAPSHOT_HEIGHT as well.',
    }),
    'auto-migrate-config': command_1.flags.boolean({
        required: false,
        env: 'AUTO_MIGRATE_CONFIG',
        description: 'Migrate user configuration automatically. Defaults to false.',
        default: false,
    }),
    'auto-start-lisk-core-v4': command_1.flags.boolean({
        required: false,
        env: 'AUTO_START_LISK_CORE',
        description: 'Start Lisk Core v4 automatically. Defaults to false. When using this flag, kindly open another terminal window to stop Lisk Core v3.1.x for when the migrator prompts.',
        default: false,
    }),
    'page-size': command_1.flags.integer({
        char: 'p',
        required: false,
        default: 100000,
        description: 'Maximum number of blocks to be iterated at once for computation. Defaults to 100000.',
    }),
    'snapshot-path': command_1.flags.string({
        required: false,
        env: 'SNAPSHOT_PATH',
        description: 'Local filepath to the state snapshot to run the migration offline. It could point either to a directory or a tarball (tar.gz).',
        dependsOn: ['network'],
        exclusive: ['snapshot-url'],
    }),
    'snapshot-url': command_1.flags.string({
        required: false,
        env: 'SNAPSHOT_URL',
        description: 'URL to download the state snapshot from. Use to run the migration offline. URL must end with tar.gz.',
        dependsOn: ['network'],
        exclusive: ['snapshot-path'],
    }),
    network: command_1.flags.enum({
        char: 'n',
        required: false,
        env: 'NETWORK',
        description: "Network to be considered for the migration. Depends on the '--snapshot-path' flag.",
        options: ['mainnet', 'testnet'],
    }),
};
module.exports = LiskMigrator;
//# sourceMappingURL=index.js.map